from .sea_result import SeaResult
from .token_container import TokenContainer
from .smart_link import SmartLink
from .base_item import BaseItem
from .file_item import FileItem
from .dir_item import DirectoryItem
from .file_item_detail import FileItemDetail
from .repo_item import RepoItem
from .dir_item_detail import DirectoryItemDetail
from .uploaded_file_item import UploadedFileItem
from .upload_file import UploadFile
from .error import Error
from .search_result_item import SearchResultItem
from .search_result import SearchResult



from datetime import datetime
from pydantic import BaseModel


class DirectoryItemDetail(BaseModel):
    """Model with detail information about the seafile directory"""
    repo_id: str
    name: str
    mtime: datetime
    path: str



from typing import Dict


class QueryParams:
    """Query parameters builder"""

    def __init__(self):
        self._query_params: Dict[str, str | int] = dict()

    def reset(self):
        self._query_params = dict()

    def get_result(self):
        if not self._query_params:
            return None

        result = self._query_params
        self.reset()
        return result

    def add_param(self, key: str, value: str | int):
        self._query_params[key] = value

    def add_param_if_exists(self, key: str, value: str | int | None):
        if value is not None:
            self.add_param(key, value)



from datetime import datetime
from .base_item import BaseItem


class FileItemDetail(BaseItem):
    """Model with detail information about the seafile file"""
    size: int
    is_draft: bool
    has_draft: bool
    can_edit: bool
    draft_id: str | None
    draft_file_path: str
    starred: bool
    permission: str
    comment_total: int
    last_modified: datetime
    last_modifier_name: str
    last_modifier_email: str
    last_modifier_contact_email: str



class UnauthorizedError(Exception):
    """Custom exception that is raised when a token obtained error occurs"""

    def __init__(self, *args: object):
        super().__init__(*args)

        if args is not None and len(args) > 0:
            self.message = args[0] if args[0] is not None else 'The access token was not obtained'



import aiohttp
from typing import Dict, List, BinaryIO, Any
from urllib.parse import urljoin
from .enums import *
from .models import *
from .builders import QueryParams
from .route_storage import RouteStorage
from .http_handlers import HttpRequestHandler, HttpDownloadHandler


class SeafileHttpClient:
    """Httpclient providing seafile web api methods."""

    def __init__(self, base_url: str):
        self._version = 'v2.1'
        self._token = None
        self._base_url = base_url
        self._route_storage = RouteStorage()

    @property
    def version(self):
        """Web api version (currently supported only v2.1)"""
        return self._version

    @property
    def base_url(self):
        """Seafile base url"""
        return self._base_url

    @property
    def token(self) -> str | None:
        """Access token"""
        return self._token

    async def ping(self):
        """Ping seafile service

        :return: SeaResult object with "pong" response
        """
        method_url = urljoin(self.base_url, self._route_storage.ping)

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url
        )
        return await handler.execute(content_type=str)

    async def auth_ping(self, token: str | None = None):
        """Ping seafile service with access token.

        :param token: access token
        :return: SeaResult object with "pong" response
        """
        method_url = urljoin(self.base_url, self._route_storage.auth_ping)

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token
        )

        return await handler.execute(content_type=str)

    async def obtain_auth_token(self, username: str, password: str):
        """Obtain access token

        :param username: seafile login
        :param password: seafile password
        :returns: SeaResult object with access token
        """
        method_url = urljoin(self.base_url, self._route_storage.auth_token)

        data = aiohttp.FormData()
        data.add_field('username', username)
        data.add_field('password', password)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=method_url,
            data=data
        )

        return await handler.execute(content_type=TokenContainer)

    async def authorize(self, username: str, password: str):
        """Authorize in seafile and save access token in context

        :param username: seafile login
        :param password: seafile password
        """
        result = await self.obtain_auth_token(username, password)

        if not result.success:
            raise Exception(
                'Error when obtain the token: ' + ', '.join(
                    e.message for e in result.errors or [Error(title='unknown', message='unknown error')]))

        self._token = result.content.token

    async def get_default_repo(self, token: str | None = None):
        """Get repository id for default repository

        :param token: access token
        :returns: SeaResult object with repository id
        """
        method_url = urljoin(self.base_url, self._route_storage.default_repo)

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token
        )

        response = await handler.execute(content_type=Dict[str, Any])
        result = SeaResult[str](
            success=response.success,
            status=response.status,
            errors=response.errors,
            content=None
        )

        if not result.success:
            return result

        if response.content:
            result.content = response.content['repo_id']
        return result

    async def create_default_repo(self, token: str | None = None):
        """Create default repository if it doesn't exist

        :param token: access token
        :returns: SeaResult object with repository id
        """
        method_url = urljoin(self.base_url, self._route_storage.default_repo)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=method_url,
            token=token or self.token
        )

        response = await handler.execute(content_type=Dict[str, Any])
        result = SeaResult[str](
            success=response.success,
            status=response.status,
            errors=response.errors,
            content=None
        )

        if not result.success:
            return result

        if response.content:
            result.content = response.content['repo_id']

        return result

    async def get_repos(self, repo_t: RepoType | None = None, token: str | None = None):
        """Get list of repos

        :param repo_t: repository types to be returned (mine, shared, group, org)
        :param token: access token
        :returns: SeaResult object with list of RepoItem
        """
        method_url = urljoin(self.base_url, self._route_storage.repos)

        query_params = QueryParams()
        query_params.add_param_if_exists('type', repo_t)

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=List[RepoItem])

    async def create_repo(self, repo_name: str, token: str | None = None) -> SeaResult[RepoItem]:
        """Create repo

        :param repo_name: name of repository being created
        :param token: access token
        :returns: SeaResult object with RepoItem that will be created
        """
        method_url = urljoin(self.base_url, self._route_storage.repos)

        data = aiohttp.FormData()
        data.add_field('name', repo_name)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=method_url,
            token=token or self.token,
            data=data
        )

        response = await handler.execute(content_type=Dict[str, Any])
        result = SeaResult[RepoItem](
            success=response.success,
            status=response.status,
            errors=response.errors,
            content=None
        )

        if not result.success:
            return result

        if response.content:
            result.content = RepoItem(
                id=response.content['repo_id'],
                type=ItemType.REPOSITORY,
                name=response.content['repo_name'],
                mtime=response.content['mtime'],
                permission=response.content['permission'],
                size=response.content['repo_size'],
                owner=response.content['email']
            )

        return result

    async def delete_repo(self, repo_id: str, token: str | None = None):
        """Delete repo

        :param repo_id: id of repository to be deleted
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.repo(repo_id))

        handler = HttpRequestHandler(
            method=HttpMethod.DELETE,
            url=method_url,
            token=token or self.token
        )

        return await handler.execute()

    async def get_upload_link(self, repo_id: str, dir_path: str, token: str | None = None):
        """Get a link to upload file

        :param repo_id: id of repository where file will be uploaded
        :param dir_path: path to the directory where file should be uploaded
        :param token: access token
        :returns: SeaResult object with upload link
        """
        method_url = urljoin(self.base_url, self._route_storage.get_upload_link(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', dir_path)

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result(),
        )

        return await handler.execute(content_type=str)

    async def upload(
            self,
            repo_id: str,
            dir_path: str,
            filename: str,
            payload: BinaryIO,
            replace: bool = False,
            relative_path: str | None = None,
            token: str | None = None
    ):
        """Upload file

        :param repo_id: id of repository where file will be uploaded
        :param dir_path: path to directory where file will be uploaded
        :param filename: name of uploaded file
        :param payload: file contents
        :param replace: indicates whether file should be overwritten if it already exists
        :param relative_path: sub-folder of "parent_dir", if this sub-folder does not exist, Seafile will create it recursively
        :param token: access token
        :returns: SeaResult object with UploadedFileItem
        """
        upload_ilnk_response = await self.get_upload_link(repo_id, dir_path, token)

        if not upload_ilnk_response.success:
            return SeaResult[UploadedFileItem](
                success=upload_ilnk_response.success,
                status=upload_ilnk_response.status,
                errors=upload_ilnk_response.errors,
                content=None
            )

        query_params = QueryParams()
        query_params.add_param('ret-json', 1)

        data = aiohttp.FormData()
        data.add_field('file', payload, filename=filename)
        data.add_field('parent_dir', dir_path)
        data.add_field('replace', str(int(replace)))

        if relative_path is not None:
            data.add_field('relative_path', relative_path)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=upload_ilnk_response.content,
            token=token or self.token,
            query_params=query_params.get_result(),
            data=data
        )

        upload_response = await handler.execute(content_type=List[UploadedFileItem])
        result = SeaResult[UploadedFileItem](
            success=upload_response.success,
            status=upload_response.status,
            errors=upload_response.errors,
            content=None
        )

        if result.success and upload_response.content is not None:
            result.content = upload_response.content.pop()

        return result

    async def uploads(
            self,
            repo_id: str,
            dir_path: str,
            files: List[UploadFile],
            replace: bool = False,
            relative_path: str | None = None,
            token: str | None = None):
        """Upload multiple files

        :param repo_id: id of repository where files will be uploaded
        :param dir_path: path to directory where files will be uploaded
        :param files: list of named tuples from file name and its contents
        :param replace: indicates whether the file should be overwritten if it already exists
        :param relative_path: sub-folder of "parent_dir", if this sub-folder does not exist, Seafile will create it recursively
        :param token: access token
        :returns: SeaResult object with list of UploadedFileItem
        """
        upload_ilnk_response = await self.get_upload_link(repo_id, dir_path, token)

        if not upload_ilnk_response.success:
            SeaResult[List[UploadedFileItem]](
                success=upload_ilnk_response.success,
                status=upload_ilnk_response.status,
                errors=upload_ilnk_response.errors,
                content=[]
            )

        query_params = QueryParams()
        query_params.add_param('ret-json', 1)

        data = aiohttp.FormData()
        data.add_field('parent_dir', dir_path)
        data.add_field('replace', str(int(replace)))
        for file in files:
            data.add_field('file', file['payload'], filename=file['filename'])

        if relative_path is not None:
            data.add_field('relative_path', relative_path)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=upload_ilnk_response.content,
            token=token or self.token,
            query_params=query_params.get_result(),
            data=data
        )

        return await handler.execute(content_type=List[UploadedFileItem])

    async def get_download_link(self, repo_id: str, filepath: str, reuse: bool = False, token: str | None = None):
        """Get a link to download file

        :param repo_id: id of repository to download file from
        :param filepath: path to file to download
        :param reuse: indicates of reuse. set it to True if you want the generated download link to be accessed more than once within one hour.
        :param token: access token
        :returns: SeaResult object with link to download file
        """
        method_url = urljoin(self.base_url, self._route_storage.file(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', filepath)
        query_params.add_param('reuse', int(reuse))

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=str)

    async def download(self, repo_id, filepath: str, token: str | None = None):
        """Download file

        :param repo_id: id of repository to download file from
        :param filepath: path to file to download
        :param token: access token
        """
        response = await self.get_download_link(repo_id, filepath, token=token)

        if not response.success:
            return SeaResult[bytes](
                success=response.success,
                status=response.status,
                errors=response.errors,
                content=None
            )

        handler = HttpDownloadHandler(
            method=HttpMethod.GET,
            url=response.content,
            token=token or self.token
        )

        return await handler.execute()

    async def get_file_detail(self, repo_id: str, filepath: str, token: str | None = None):
        """Get detail information about the file

        :param repo_id: id of repository where file is located
        :param filepath: path to file
        :param token: access token
        :returns: SeaResult object with FileItemDetail
        """
        method_url = urljoin(self.base_url, self._route_storage.file_detail(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', filepath)

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=FileItemDetail)

    async def create_file(self, repo_id: str, filepath: str, token: str | None = None):
        """Create new file

        :param repo_id: id of repository to create file in
        :param filepath: path to new file
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.file(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', filepath)

        data = aiohttp.FormData()
        data.add_field('operation', FileOperation.CREATE)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result(),
            data=data
        )

        return await handler.execute()

    async def rename_file(self, repo_id: str, filepath: str, new_filename: str, token: str | None = None):
        """Rename file

        :param repo_id: id of repository where file is located
        :param filepath: path to file
        :param new_filename: new name of file
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.file(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', filepath)

        data = aiohttp.FormData()
        data.add_field('operation', FileOperation.RENAME)
        data.add_field('newname', new_filename)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result(),
            data=data
        )

        return await handler.execute()

    async def move_file(
            self,
            repo_id: str,
            filepath: str,
            dst_dir: str,
            token: str | None = None,
            dst_repo_id: str | None = None):
        """Move file to another location

        :param repo_id: id of repository where file is located
        :param filepath: path to file
        :param dst_dir: directory where file will be moved
        :param dst_repo_id: id of repository where file will be moved
        :param token: access token
        :returns: SeaResult object with a link to new file location
        """
        method_url = urljoin(self.base_url, self._route_storage.file(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', filepath)

        data = aiohttp.FormData()
        data.add_field('operation', FileOperation.MOVE)
        data.add_field('dst_dir', dst_dir)
        data.add_field('dst_repo', dst_repo_id or repo_id)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result(),
            data=data
        )

        return await handler.execute(content_type=str)

    async def copy_file(
            self,
            repo_id: str,
            filepath: str,
            dst_dir: str,
            token: str | None = None,
            dst_repo_id: str | None = None):
        """Copy file

        :param repo_id: id of repository where file is located
        :param filepath: path to file
        :param dst_dir: directory where a copy of file will be created
        :param dst_repo_id: id of repository where a copy of file will be created
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.file(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', filepath)

        data = aiohttp.FormData()
        data.add_field('operation', FileOperation.COPY)
        data.add_field('dst_dir', dst_dir)
        data.add_field('dst_repo', dst_repo_id or repo_id)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result(),
            data=data
        )

        return await handler.execute()

    async def delete_file(self, repo_id: str, filepath: str, token: str | None = None):
        """Delete file

        :param repo_id: id of repository where file is located
        :param filepath: path to file to be deleted
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.file(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', filepath)

        data = aiohttp.FormData()
        data.add_field('operation', FileOperation.DELETE)

        handler = HttpRequestHandler(
            method=HttpMethod.DELETE,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result(),
            data=data
        )

        return await handler.execute()

    async def lock_file(self, repo_id: str, filepath: str, token: str | None = None):
        """Lock file

        :param repo_id: id of repository where file is located
        :param filepath: path to file
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.file(repo_id))

        data = aiohttp.FormData()
        data.add_field('p', filepath)
        data.add_field('operation', FileOperation.LOCK)

        handler = HttpRequestHandler(
            method=HttpMethod.PUT,
            url=method_url,
            token=token or self.token,
            data=data
        )

        return await handler.execute()

    async def unlock_file(self, repo_id: str, filepath: str, token: str | None = None):
        """Unlock file

        :param repo_id: id of repository where file is located
        :param filepath: path to file
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.file(repo_id))

        data = aiohttp.FormData()
        data.add_field('p', filepath)
        data.add_field('operation', FileOperation.UNLOCK)

        handler = HttpRequestHandler(
            method=HttpMethod.PUT,
            url=method_url,
            token=token or self.token,
            data=data
        )

        return await handler.execute()

    async def get_items(self, repo_id: str, path: str | None = None, token: str | None = None):
        """Get all items in a directory

        :param repo_id: id of repository to get information from
        :param path: path to directory where you need to find out what is located
        :param token: access token
        :returns: SeaResult object with list of BaseItem
        """
        method_url = urljoin(self.base_url, self._route_storage.dir(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', path or '/')

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=List[BaseItem])

    async def get_items_by_id(self, repo_id: str, dir_id: str, token: str | None = None):
        """Get all items in a directory by directory id

        :param repo_id: id of repository to get information from
        :param dir_id: id of directory where you need to find out what is located
        :param token: access token
        :returns: SeaResult object with list of BaseItem
        """
        method_url = urljoin(self.base_url, self._route_storage.dir(repo_id))

        query_params = QueryParams()
        query_params.add_param('oid', dir_id)

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=List[BaseItem])

    async def get_files(self, repo_id: str, path: str | None = None, token: str | None = None):
        """Get all files in a directory

        :param repo_id: id of repository to get information from
        :param path: path to directory where you need to find out what is located
        :param token: access token
        :returns: SeaResult object with list of FileItem
        """
        method_url = urljoin(self.base_url, self._route_storage.dir(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', path or '/')
        query_params.add_param('t', 'f')

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=List[FileItem])

    async def get_files_by_id(self, repo_id: str, dir_id: str, token: str | None = None):
        """Get all files in a directory by directory id

        :param repo_id: id of repository to get information from
        :param dir_id: id of directory where you need to find out what is located
        :param token: access token
        :returns: SeaResult object with list of FileItem
        """
        method_url = urljoin(self.base_url, self._route_storage.dir(repo_id))

        query_params = QueryParams()
        query_params.add_param('oid', dir_id)
        query_params.add_param('t', 'f')

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=List[FileItem])

    async def get_directories(
            self,
            repo_id: str,
            path: str | None = None,
            recursive: bool = False,
            token: str | None = None):
        """Get all directories in a directory

        :param repo_id: id of repository to get information from
        :param path: path to directory where you need to find out what is located
        :param recursive: indicates a recursive search method
        :param token: access token
        :returns: SeaResult object with list of DirectoryItem
        """
        method_url = urljoin(self.base_url, self._route_storage.dir(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', path or '/')
        query_params.add_param('t', 'd')
        query_params.add_param('recursive', int(recursive))

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=List[DirectoryItem])

    async def get_directories_by_id(
            self,
            repo_id: str,
            dir_id: str,
            recursive: bool = False,
            token: str | None = None):
        """Get all directories in a directory by directory id

        :param repo_id: id of repository to get information from
        :param dir_id: id of directory where you need to find out what is located
        :param recursive: indicates a recursive search method
        :param token: access token
        :returns: SeaResult object with list of DirectoryItem
        """
        method_url = urljoin(self.base_url, self._route_storage.dir(repo_id))

        query_params = QueryParams()
        query_params.add_param('oid', dir_id)
        query_params.add_param('t', 'd')
        query_params.add_param('recursive', int(recursive))

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=List[DirectoryItem])

    async def get_directory_detail(self, repo_id: str, path: str, token: str | None = None):
        """Get detailed information about the directory

        :param repo_id: id of repository where directory is located
        :param path: path to directory
        :param token: access token
        :returns: SeaResult object with DirectoryItemDetail
        """
        if path == '/':
            raise ValueError('Path should not be "/"')

        method_url = urljoin(self.base_url, self._route_storage.dir_detail(repo_id))

        query_params = QueryParams()
        query_params.add_param('path', path)

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=DirectoryItemDetail)

    async def create_directory(self, repo_id: str, path: str, token: str | None = None):
        """Create new directory

        :param repo_id: id of repository where directory will be created
        :param path: path to directory
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.dir(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', path)

        data = aiohttp.FormData()
        data.add_field('operation', DirectoryOperation.CREATE)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=method_url,
            token=token or self.token,
            data=data,
            query_params=query_params.get_result()
        )

        return await handler.execute()

    async def rename_directory(self, repo_id: str, path: str, new_name: str, token: str | None = None):
        """Rename directory

        :param repo_id: id of repository where directory is located
        :param path: path to directory
        :param new_name: new name of directory
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.dir(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', path)

        data = aiohttp.FormData()
        data.add_field('operation', DirectoryOperation.RENAME)
        data.add_field('newname', new_name)

        handler = HttpRequestHandler(
            method=HttpMethod.POST,
            url=method_url,
            token=token or self.token,
            data=data,
            query_params=query_params.get_result()
        )

        return await handler.execute()

    async def delete_directory(self, repo_id: str, path: str, token: str | None = None):
        """Delete directory

        :param repo_id: id of repository where directory will be deleted
        :param path: path to directory
        :param token: access token
        """
        method_url = urljoin(self.base_url, self._route_storage.dir(repo_id))

        query_params = QueryParams()
        query_params.add_param('p', path)

        handler = HttpRequestHandler(
            method=HttpMethod.DELETE,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute()

    async def get_smart_link(self, repo_id: str, path: str, is_dir: bool = False, token: str | None = None):
        """Get smart link to item

        :param repo_id: id of repository where item is located
        :param path: path to item
        :param is_dir: indicates that the item is a directory
        :param token: access token
        :returns: SeaResult with smart-link
        """
        method_url = urljoin(self.base_url, self._route_storage.smart_link)

        query_params = QueryParams()
        query_params.add_param('repo_id', repo_id)
        query_params.add_param('path', path)
        query_params.add_param('is_dir', str(is_dir).lower())

        handler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        return await handler.execute(content_type=SmartLink)

    async def search_file(
            self,
            query: str,
            repo_id: str,
            token: str | None = None
    ):
        """Search files in repositories

        :param query: keyword for searching
        :param repo_id: id of repository where search will be performed
        :param token: access token
        :returns: SeaResult with list of SearchResultItem
        """
        method_url = urljoin(self.base_url, self._route_storage.search_file)

        query_params = QueryParams()
        query_params.add_param('q', query)
        query_params.add_param('repo_id', repo_id)

        hanndler = HttpRequestHandler(
            method=HttpMethod.GET,
            url=method_url,
            token=token or self.token,
            query_params=query_params.get_result()
        )

        response = await hanndler.execute(content_type=SearchResult)
        result = SeaResult[List[SearchResultItem]](
            success=response.success,
            status=response.status,
            errors=response.errors,
            content=None
        )

        if result.success and response.content is not None:
            result.content = response.content.data

        return result



from http import HTTPStatus
from typing import Dict, Any
from pydantic import parse_raw_as
from ..enums import HttpMethod
from abc import ABCMeta, abstractmethod
from ..exceptions import UnauthorizedError
from ..models import SeaResult, Error


class BaseHttpHandler(metaclass=ABCMeta):
    SUCCESS_STATUSES = (
        HTTPStatus.OK,
        HTTPStatus.CREATED,
        HTTPStatus.ACCEPTED,
        HTTPStatus.NON_AUTHORITATIVE_INFORMATION,
        HTTPStatus.NO_CONTENT,
        HTTPStatus.RESET_CONTENT,
        HTTPStatus.PARTIAL_CONTENT,
        HTTPStatus.MULTI_STATUS,
        HTTPStatus.ALREADY_REPORTED,
        HTTPStatus.IM_USED
    )

    def __init__(
            self,
            method: HttpMethod,
            url: str,
            token: str | None = None,
            headers: Dict[str, str] | None = None,
            query_params: Dict[str, str | int] | None = None,
            data: Any | None = None):
        self._method = method
        self._route = url
        self._data = data
        self._token = token
        self._query_params = query_params
        self._headers: Dict[str, str] = dict()
        if token is not None:
            self._headers |= self._create_authorization_headers(token)
        if headers is not None:
            self._headers |= headers

    @abstractmethod
    async def execute(self, *args, **kwargs) -> SeaResult:
        ...

    @staticmethod
    def _try_parse_errors(response_content: str | bytes):
        try:
            errors = parse_raw_as(Dict[str, Any], response_content)
            result = list()
            for key, value in errors.items():
                if isinstance(value, list):
                    result.append(Error(title=key, message='; '.join(value)))
                else:
                    result.append(Error(title=key, message=value))

            return result
        except Exception:
            pass

    @staticmethod
    def _create_authorization_headers(token: str | None):
        if token is None:
            raise UnauthorizedError('Undefined token')

        return {'Authorization': f'Token {token}'}



from .base import StrEnum


class HttpMethod(StrEnum):
    """Enumeration of popular http methods"""
    GET = 'get'
    POST = 'post'
    PUT = 'put'
    DELETE = 'delete'
    HEAD = 'head'
    PATCH = 'patch'



from ..enums import ItemType
from pydantic import BaseModel


class BaseItem(BaseModel):
    """Model with information about the seafile item (repo, file, directory ...)"""
    id: str
    type: ItemType
    name: str
    mtime: int
    permission: str



from .base import StrEnum


class FileOperation(StrEnum):
    """Enumeration of operations that can be performed on files"""

    CREATE = 'create'

    RENAME = 'rename'

    LOCK = 'lock'

    UNLOCK = 'unlock'

    MOVE = 'move'

    COPY = 'copy'

    DELETE = 'delete'



from .base_item import BaseItem


class FileItem(BaseItem):
    """Model with information about the seafile file"""
    size: int
    modifier_name: str
    modifier_email: str
    modifier_contact_email: str
    starred: bool



from enum import Enum, auto


class StrEnum(str, Enum):
    """
    StrEnum is a Python ``enum.Enum`` that inherits from ``str``. The default
    ``auto()`` behavior uses the member name as its value.

    Example usage::

        class Example(StrEnum):
            UPPER_CASE = auto()
            lower_case = auto()
            MixedCase = auto()

        assert Example.UPPER_CASE == "UPPER_CASE"
        assert Example.lower_case == "lower_case"
        assert Example.MixedCase == "MixedCase"
    """

    def __new__(cls, value, *args, **kwargs):
        if not isinstance(value, (str, auto)):
            raise TypeError(
                f"Values of StrEnums must be strings: {value!r} is a {type(value)}"
            )
        return super().__new__(cls, value, *args, **kwargs)

    def __str__(self):
        return str(self.value)

    def _generate_next_value_(self, *_):
        return self



from .http_client import SeafileHttpClient
from .models import (
    SeaResult,
    BaseItem,
    DirectoryItem,
    DirectoryItemDetail,
    FileItem,
    FileItemDetail,
    UploadFile,
    UploadedFileItem,
    RepoItem,
    SmartLink
)

from .enums import (
    ItemType,
    RepoType
)



class RouteStorage:
    """Route storage for storing, processing and giving routes of seafile web api"""

    DEFAULT_SUFFIX = 'api2/'

    PING_ROUTE = 'ping/'
    AUTH_PING_ROUTE = 'auth/ping'
    AUTH_TOKEN_ROUTE = 'auth-token/'
    REPO_ROUTE = 'repos/'
    DEFAULT_REPO_ROUTE = 'default-repo/'
    DIR_ROUTE = 'repos/{repo_id}/dir/'
    FILE_ROUTE = 'repos/{repo_id}/file/'
    SMART_LINK_ROUTE = 'smart-link/'
    GET_UPLOAD_LINK_ROUTE = 'repos/{repo_id}/upload-link/'
    SEARCH_ROUTE = 'search-file/'

    def __init__(self, version: str = 'v2.1', suffix: str | None = None):
        self._version = version
        self._suffix = suffix or self.DEFAULT_SUFFIX

    @property
    def ping(self):
        return self._suffix + self.PING_ROUTE

    @property
    def auth_ping(self):
        return self._suffix + self.AUTH_PING_ROUTE

    @property
    def auth_token(self):
        return self._suffix + self.AUTH_TOKEN_ROUTE

    @property
    def repos(self):
        return self._suffix + self.REPO_ROUTE

    @property
    def default_repo(self):
        return self._suffix + self.DEFAULT_REPO_ROUTE

    @property
    def smart_link(self):
        return 'api/' + self._version + '/' + self.SMART_LINK_ROUTE

    @property
    def search_file(self):
        return 'api/' + self._version + '/' + self.SEARCH_ROUTE

    def repo(self, repo_id: str):
        return self._suffix + self.REPO_ROUTE + repo_id + '/'

    def file(self, repo_id: str):
        return self._suffix + self.FILE_ROUTE.format(repo_id=repo_id)

    def file_detail(self, repo_id: str):
        return self._suffix + self.FILE_ROUTE.format(repo_id=repo_id) + 'detail/'

    def dir(self, repo_id: str):
        return self._suffix + self.DIR_ROUTE.format(repo_id=repo_id)

    def dir_detail(self, repo_id: str):
        return 'api/' + self._version + '/' + self.DIR_ROUTE.format(repo_id=repo_id) + 'detail/'

    def get_upload_link(self, repo_id: str):
        return self._suffix + self.GET_UPLOAD_LINK_ROUTE.format(repo_id=repo_id)



import aiohttp
from http import HTTPStatus
from typing import Dict, Any
from .base_http_handler import BaseHttpHandler
from ..enums import HttpMethod
from ..models import SeaResult


class HttpDownloadHandler(BaseHttpHandler):

    def __init__(
            self,
            method: HttpMethod,
            url: str,
            token: str | None = None,
            headers: Dict[str, str] | None = None,
            query_params: Dict[str, str | int] | None = None,
            data: Any | None = None):
        super().__init__(method, url, token, headers, query_params, data)

    async def execute(self) -> SeaResult[bytes]:
        async with aiohttp.ClientSession() as session:
            async with session.request(
                    method=self._method,
                    url=self._route,
                    headers=self._headers,
                    params=self._query_params,
                    data=self._data
            ) as response:
                response_content = await response.content.read()
                http_status = HTTPStatus(response.status)

                result = SeaResult[bytes](
                    success=(http_status in self.SUCCESS_STATUSES),
                    status=http_status,
                    errors=None,
                    content=None
                )

                if result.success:
                    result.content = response_content
                else:
                    result.errors = self._try_parse_errors(response_content)

                return result



from .error import Error
from typing import List, TypeVar, Generic
from pydantic.generics import GenericModel
from http import HTTPStatus

# Type of method execution result
ContentT = TypeVar('ContentT')


class SeaResult(GenericModel, Generic[ContentT]):
    """Model for providing information about the result of executing an APi method"""

    # Indicates the success of the method execution
    success: bool

    # Http status
    status: HTTPStatus

    # List of errors
    errors: List[Error] | None

    # Result of method execution
    content: ContentT | None

    class Config:
        arbitrary_types_allowed = True
